"""
Module realizes background plugin example
"""
import asyncio
from asyncio import Task
from typing import Optional

from luna_plugins.base.plugins_meta.base_plugins import BaseBackgroundPlugin
from sanic.response import HTTPResponse

from ...web.handlers import BaseHandler


class HandlerExample(BaseHandler):
    """
    Handler example
    """

    async def get(self) -> HTTPResponse:
        """
        Method get example.

        Returns:
            response
        """
        return self.success(body="I am teapot", contentType="text/plain", statusCode=418)

    @property
    def app(self):
        """
        Get app
        Abstract method of `BaseHandler`
        Returns:
            app
        """
        return self.request.app

    @property
    def config(self):
        """
        Get app config
        Abstract method of `BaseHandler`
        Returns:
            app config
        """
        return self.request.app.ctx.serviceConfig


class BackgroundPluginExample(BaseBackgroundPlugin):
    """
    Background plugin example.

    Create background task and add a route.
    """

    def __init__(self, app: "LunaApplication"):
        super().__init__(app)
        app.router.reset()
        app.addRoutes([("/teapot", HandlerExample)])
        app.finalize()
        self.task: Optional[Task] = None
        self.temperature = 0

    async def initialize(self):
        """
        Initialize plugin
        """

    async def close(self):
        """
        Stop background process
        Returns:

        """
        if self.task:
            self.task.cancel()

    async def usefulJob(self):
        """
        Some useful  async work
        """
        while True:
            await asyncio.sleep(1)
            self.temperature = min(100, self.temperature + 1)
            if self.temperature < 100:
                self.app.ctx.logger.info(f"I boil water, temperature: {self.temperature}")
            else:
                self.app.ctx.logger.info("boiling water is ready, would you care for a cup of tea?")

    async def start(self):
        """
        Run background process

        .. warning::
            The function suppose that the process is handle in this coroutine. The coroutine must start
            the process only without awaiting a end of the process
        """
        self.task = asyncio.create_task(self.usefulJob())
