"""
Module contains platform-specific types
"""
from datetime import datetime
from typing import Any, Optional, Type, Union

import fastjsonschema
from ciso8601 import parse_datetime  # pylint: disable-msg=no-name-in-module
from dateutil import tz
from dateutil.relativedelta import relativedelta
from pydantic import Field
from pydantic.types import (  # pylint: disable-msg=E0611
    ConstrainedFloat as _ConstrainedFloat,
    ConstrainedStr as _ConstrainedStr,
    OptionalIntFloat,
    StrictStr as _StrictStr,
    confloat as _confloat,
    conint as _conint,
    constr as _constr,
)

# pylint: disable-msg=no-name-in-module
from pydantic.validators import float_validator, number_multiple_validator, number_size_validator
from vlutils.helpers import UTC_TIMEZONE
from vlutils.regexps import UUID_REGEXP_STR

from ..utils.regexps import VALID_DB_STRING_PATTERN

ObjectId = _constr(strict=True, regex=UUID_REGEXP_STR)
TrackId = _constr(strict=True, regex=r"^[a-zA-Z0-9_\-]{1,36}$")
DATETIME_SCHEMA = fastjsonschema.compile({"type": "string", "format": "date-time"})
DATETIME_NOW_SCHEMA = fastjsonschema.compile({"type": "string", "pattern": r"^now-(\d+)[smhdwMy]$"})


def OptionalNotNullable() -> object:  # pylint: disable-msg=invalid-name
    """
    Pydantic hack for not required not nullable fields. If field does not present in an input json will be
    initialize of the None, if field is present as null in the input json will be raise corresponding error.
    Examples:
        >>> class A(BaseModel):
        >>>     id: int
        >>>     description: str = OptionalNotNullable()
        if input json is equal to `{"id": 1, "description": None}` will be raize the error.
        if input json is equal to `{"id": 1}` will be correct parsing.
    Returns:
        Field with default None
    """
    return Field(default_factory=lambda: None)


def Int(minValue: Optional[int] = None, maxValue: Optional[int] = None):  # pylint: disable-msg=invalid-name
    """Get pydantic int model with min/max length"""
    return _conint(ge=minValue, le=maxValue)


def Float(minValue: Optional[float] = None, maxValue: Optional[float] = None):  # pylint: disable-msg=invalid-name
    """Get pydantic float model with min/max length"""
    return _confloat(ge=minValue, le=maxValue)


def Str(minLength: Optional[int] = None, maxLength: Optional[int] = None):  # pylint: disable-msg=invalid-name
    """Get pydantic strict string model with min/max length"""
    return _constr(strict=True, regex=VALID_DB_STRING_PATTERN, min_length=minLength, max_length=maxLength)


def QueryStr(minLength: Optional[int] = None, maxLength: Optional[int] = None):  # pylint: disable-msg=invalid-name
    """Get pydantic strict string model with min/max length"""
    return _constr(strict=False, regex=VALID_DB_STRING_PATTERN, min_length=minLength, max_length=maxLength)


class BoolFromInt01(_ConstrainedStr):
    """Bool from 0/1 int getter"""

    @classmethod
    def validate(cls, value: str) -> bool:  # type: ignore
        """Validate value and morph to bool"""
        value = int(value)  # type: ignore
        if value not in (0, 1):
            raise ValueError
        return bool(value)


class DatetimeString(_StrictStr):
    """Datetime string in rfc3339 format."""

    @classmethod
    def validate(cls, value: str) -> str:
        """
        Validate that value can be parsed as time
        Args:
            value: time in rfc3339
        Returns:
            value
        """
        value = super().validate(value)  # type: ignore
        parse_datetime(value).astimezone(UTC_TIMEZONE)
        return value


class Latitude(_ConstrainedFloat):
    """Non strict float latitude"""

    ge = -90.0
    le = 90.0


class Longitude(_ConstrainedFloat):
    """Non strict float longitude"""

    ge = -180.0
    le = 180.0


class ValidString(_ConstrainedStr):
    """String validator."""

    @classmethod
    def validate(cls, value: str) -> str:
        """Validate string."""

        if VALID_DB_STRING_PATTERN in value:
            raise ValueError
        return value


class ListStringsValidator(ValidString):
    """List string validator."""

    @classmethod
    def validate(cls, value: str) -> list[str]:
        """Create list with values from string."""

        return value.split(",")


class _CustomFloatMeta(type):
    """Custom float meta"""

    def __getitem__(cls, param: Any) -> type:
        return type("CustomFloat", (CustomFloat,), {"param": param})


class CustomFloat(metaclass=_CustomFloatMeta):
    """
    Custom float - same as default pydantic non-strict float, but not allow to use string as float
    """

    gt: OptionalIntFloat = None
    ge: OptionalIntFloat = None
    lt: OptionalIntFloat = None
    le: OptionalIntFloat = None
    param: Any = None
    multiple_of: OptionalIntFloat = None

    @classmethod
    def __get_validators__(cls):
        """Get validators - same as pydantic default, but with non-string validation"""
        yield cls.validateNonStr
        yield float_validator
        yield number_size_validator
        yield number_multiple_validator

    @classmethod
    def validateNonStr(cls, value) -> float:
        """Validate value is not string"""
        if isinstance(value, str):
            raise ValueError("value is not a valid float")
        return value


def customfloat(*, gt: int = None, ge: int = None, lt: int = None, le: int = None) -> Type[float]:
    """Construct Field for custom float."""
    return type("CustomFloatValue", (CustomFloat,), {ge: ge, lt: lt, le: le, gt: gt})


class _CustomDatetimeMeta(type):
    """Custom datetime meta"""

    def __getitem__(cls, param: Any) -> type:
        return type("CustomDatetime", (CustomDatetime,), {"param": param})


class CustomDatetime(metaclass=_CustomDatetimeMeta):
    """
    Custom extended datetime - do not use default if it needs to get attribute as str in ISO-format,
    """

    @classmethod
    def __get_validators__(cls):
        """Use own validator."""
        yield cls.validateDatetime

    @classmethod
    def validateDatetime(cls, value: str | datetime) -> str:
        """
        Validate value is not string
        Args:
            value: input value
        Returns:
            datetime in string format
        """
        try:
            DATETIME_SCHEMA(value)
            return str(value)
        except fastjsonschema.exceptions.JsonSchemaException:  # pylint: disable-msg=W0612
            raise ValueError("Invalid datetime") from None


class CustomExtendedDatetime(metaclass=_CustomDatetimeMeta):
    """
    Custom datetime - do not use default if it needs to get attribute as str in ISO-format
    """

    _storageTime: str

    @classmethod
    def initialize(cls, storageTime: str):
        """
        Initialize storage time
        Args:
            storageTime: storage time - utc or local
        """
        cls._storageTime = storageTime

    @classmethod
    def __get_validators__(cls):
        """Get validators - use pydantic default validation, but return attribute as valid datetime str in ISO-format"""
        yield cls.validateDatetime

    @classmethod
    def validateDatetime(cls, value: Union[str, datetime]) -> str:
        """
        Validate value is not string
        Args:
            value: input value
        Returns:
            datetime in string format
        """
        if isinstance(value, str) and value.startswith("now-"):
            try:
                DATETIME_NOW_SCHEMA(value)
            except fastjsonschema.exceptions.JsonSchemaException:
                raise ValueError("Invalid datetime") from None
            splitDelta = value.split("-")[1]
            deltaValue, deltaType = int(splitDelta[:-1]), splitDelta[-1]
            kwarg = dict(y="years", M="months", w="weeks", d="days", h="hours", m="minutes", s="seconds")[deltaType]
            try:
                datetimeResult = (
                    (datetime.utcnow() - relativedelta(**{kwarg: deltaValue})).isoformat() + "Z"
                    if cls._storageTime == "UTC"
                    else (
                        datetime.now().replace(tzinfo=tz.tzlocal()) - relativedelta(**{kwarg: deltaValue})
                    ).isoformat()
                )
            except ValueError:
                raise ValueError("Invalid datetime") from None
            return datetimeResult
        try:
            DATETIME_SCHEMA(value)
            return str(value)
        except fastjsonschema.exceptions.JsonSchemaException:
            raise ValueError("Invalid datetime") from None
