"""
Module realizes useful db function.
"""
import asyncio
from datetime import datetime
from functools import wraps
from typing import List

from luna_plugins.utils.vlmatch.sa import VLMatch as NewVLMatch
from sqlalchemy import text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.elements import TextClause
from vlutils.helpers import convertTimeToString, convertToDateTimeObject

from ..enums.databases import Databases
from ..errors.errors import Error
from ..errors.exception import VLException
from .log import Logger, logger

# internal database match function name
DB_MATCH_FUNCTION_NAME = "VLMatch"


def dbExceptionWrap(wrapFunc):
    """
    Decorator for catching exceptions when composing queries to the database.

    Args:
        wrapFunc: function

    Returns:
        If exception is not caught, function result is returned, else Result\
        with value of exception is returned
    """

    if asyncio.iscoroutinefunction(wrapFunc):

        @wraps(wrapFunc)
        async def wrap(*func_args, **func_kwargs):
            try:
                return await wrapFunc(*func_args, **func_kwargs)
            except (TypeError, VLException):
                raise
            except Exception as exc:
                func_args[0].logger.exception()
                raise VLException(Error.ExecuteError) from exc

    else:

        @wraps(wrapFunc)
        def wrap(*func_args, **func_kwargs):
            try:
                return wrapFunc(*func_args, **func_kwargs)
            except (TypeError, VLException):
                raise
            except Exception as exc:
                func_args[0].logger.exception()
                raise VLException(Error.ExecuteError) from exc

    return wrap


def getColumnNames(tableModel: declarative_base) -> List[str]:
    """
    Get all column names of table. <Dirty hack>.

    Returns:
        list of column name in order as in db
    """
    return tableModel.__table__.columns._data._list  # pylint: disable-msg=W0212


# raw sql function to get utc timestamp for creating/updating table columns
ORACLE_CURRENT_TIMESTAMP = text("SYS_EXTRACT_UTC(SYSTIMESTAMP)")
POSTGRES_CURRENT_TIMESTAMP = text("TIMEZONE('UTC', NOW())")


def currentDBTimestamp(dbType: str) -> TextClause:
    """
    Get expression for an estimation current UTC timestamp in database
    Args:
        dbType: database type

    Returns:
        text clause
    Raises:
        RuntimeError: if database type is unknown
    """
    if dbType == Databases.POSTGRES.value:
        return POSTGRES_CURRENT_TIMESTAMP
    if dbType == Databases.ORACLE.value:
        return ORACLE_CURRENT_TIMESTAMP
    raise RuntimeError(f"Unsupported database type {dbType}")


def convertDateTimeToStr(inputDateTime: datetime, inUTC: bool) -> str:
    """
    Function adds current time zone and isoformat(T) to datetime object.
    If there is no time zone information, UTC will be applied by default.

    Args:
        inputDateTime: datetime object
        inUTC: return time in UTC format or local
    Returns:
        string, format: "YYYY-MM-DDThh:mm:ss.sTZD" (eg 1997-07-16T19:20:30.45+01:00)"
    """
    Logger().warning(
        "Do not use function 'convertDateTimeToStr' (deprecated). Use 'convertTimeToString' from vlutils.helpers"
    )
    return convertTimeToString(inputDateTime, inUTC)


def convertToDataTimeObject(*args, **kwargs):
    """
    Time will be converted to UTC without timezone information. Deprecated.
    """
    logger.warning(
        "Do not use function 'convertToDataTimeObject' (deprecated). Use 'convertToDateTimeObject' "
        "from vlutils.helpers"
    )
    return convertToDateTimeObject(*args, **kwargs)


class VLMatch(NewVLMatch):  # pylint: disable-msg=too-many-ancestors
    """
    VLMatch database function abstraction. Deprecated.
    """

    def __init__(self, *args, **kwargs):
        logger.warning("Do not use class 'VLMatch' (deprecated). Use 'VLMatch' from luna_plugins.utils.vlmatch.sa")
        super().__init__(*args, **kwargs)
